import logging
from datetime import datetime
import random
import threading
from p2pfl.settings import Settings

###############################
#    CommunicationProtocol    # --> Invoker of Command Patern
###############################

class CommunicationProtocol:
    """
    Manages the meaning of node communication messages. Some messages contain a hash at end, it is used as a unique identifier for the message, 
    this kind of messages are gossiped to the entire network.
    
    The valid messages can be classified into gossiped and non-gossiped:
        Gossiped messages: 
            - BEAT <node> <HASH> 
            - START_LEARNING <rounds> <epoches> <HASH>
            - STOP_LEARNING <HASH>
            - VOTE_TRAIN_SET <node> (<node> <punct>)* VOTE_TRAIN_SET_CLOSE <HASH> 
            - METRICS <node> <round> <loss> <metric> <HASH> 

        Non Gossiped messages (communication over only 2 nodes):
            - CONNECT <ip> <port> <full> <force>
            - CONNECT_TO <ip> <port>
            - STOP 
            - PARAMS <data> \PARAMS
            - MODELS_READY <round> 
            - MODELS_AGGREGATED <node>* MODELS_AGGREGATED_CLOSE
            - MODEL_INITIALIZED

    Furthermore, all messages consists of encoded text (utf-8), with the exception of the `PARAMS` message, which contains serialized binaries.

    Non-static methods are used to process the different messages. Static methods are used to build messages and process only the `CONNECT` message (handshake). 
    
    Args:
        command_dict: Dictionary with the callbacks to execute at `process_message`.

    Attributes:
        command_dict: Dictionary with the callbacks to execute at `process_message`.
        last_messages: List of the last messages received.
    """

    BEAT = "BEAT"
    """
    Beat message header.
    """
    STOP = "STOP"           
    """
    Stop message header.
    """
    CONN = "CONNECT" 
    """
    Connection message header.
    """        
    CONN_TO = "CONNECT_TO"
    """
    Connection to message header.
    """
    START_LEARNING = "START_LEARNING"
    """
    Start learning message header.
    """
    STOP_LEARNING = "STOP_LEARNING"
    """
    Stop learning message header.
    """
    PARAMS = "PARAMS" # special case (binary) 
    """
    Parameters message header.
    """
    PARAMS_CLOSE = "\PARAMS" # special case (binary)
    """
    Parameters message closing.
    """
    MODELS_READY = "MODELS_READY"    
    """
    Models ready message header.
    """ 
    METRICS = "METRICS"
    """
    Metrics message header.
    """
    VOTE_TRAIN_SET = "VOTE_TRAIN_SET" 
    """
    Vote train set message header.
    """
    VOTE_TRAIN_SET_CLOSE = "\VOTE_TRAIN_SET"
    """
    Vote train set message closing.
    """
    MODELS_AGGREGATED = "MODELS_AGGREGATED"    
    """
    Models aggregated message header.
    """
    MODELS_AGGREGATED_CLOSE = "\MODELS_AGGREGATED" 
    """
    Models aggregated message closing.
    """
    MODEL_INITIALIZED = "MODEL_INITIALIZED" 
    """
    Model initialized message header.
    """
    
    ############################################
    #    MSG PROCESSING (Non Static Methods)   #
    ############################################

    def __init__(self, command_dict):
        self.command_dict = command_dict
        self.last_messages = []
        self.__last_messages_lock = threading.Lock()

    def add_processed_messages(self,messages):
        """
        Add messages to the last messages list. If ammount is higher than `Settings.AMOUNT_LAST_MESSAGES_SAVED` remove the oldest to keep the size.

        Args:
            hash_messages: List of hashes of the messages.
        """
        self.__last_messages_lock.acquire()
        self.last_messages = self.last_messages + messages
        # Remove oldest messages
        if len(self.last_messages)>Settings.AMOUNT_LAST_MESSAGES_SAVED:
            self.last_messages = self.last_messages[len(self.last_messages)-Settings.AMOUNT_LAST_MESSAGES_SAVED:]
        self.__last_messages_lock.release()

    def process_message(self, msg):
        """
        Processes messages and executes the callback associated with it (from ``command_dict``).        
        
        Args:
            msg: The message to process.

        Returns:
            tuple: (messages_executed, error) messages_executed is a list of the messages executed, error true if there was an error.

        """
        self.tmp_exec_msgs = {}
        error = False

        # Determine if is a binary message or not
        header = CommunicationProtocol.PARAMS.encode("utf-8")
        if msg[0:len(header)] == header:
            end = CommunicationProtocol.PARAMS_CLOSE.encode("utf-8")

            # Check if done
            end_pos = msg.find(end)
            if end_pos != -1:
                return [], not self.__exec(CommunicationProtocol.PARAMS, None, None, msg[len(header):end_pos], True)

            return [],not self.__exec(CommunicationProtocol.PARAMS, None, None, msg[len(header):], False)

        else:      
            # Try to decode the message
            message = ""
            try:
                message = msg.decode("utf-8")
                message = message.split()
            except:
                error = True
        
            # Process messages
            while len(message) > 0:

                # Beat
                if message[0] == CommunicationProtocol.BEAT:
                    if len(message) > 2:
                        hash_ = message[2]
                        cmd_text = (" ".join(message[0:3]) + "\n").encode("utf-8")
                        if self.__exec(CommunicationProtocol.BEAT,hash_, cmd_text, message[1]):
                            message = message[3:]
                        else:
                            error = True
                            break
                    else:
                        error = True
                        break
                    
                # Stop (non gossiped)
                elif message[0] == CommunicationProtocol.STOP:
                    if self.__exec(CommunicationProtocol.STOP, None, None):
                        message = message[1:]
                    else:
                        error = True
                        break
                   

                # Connect to
                elif message[0] == CommunicationProtocol.CONN_TO:
                    if len(message) > 2:
                        if message[2].isdigit():
                            if self.__exec(CommunicationProtocol.CONN_TO, None, None, message[1], int(message[2])):
                                message = message[3:]
                            else:
                                error = True
                                break 
                        else:
                            error = True
                            break
                    else:
                        error = True
                        break

                # Start learning
                elif message[0] == CommunicationProtocol.START_LEARNING:
                    if len(message) > 3:
                        if message[1].isdigit() and message[2].isdigit():
                            hash_ = message[3]
                            cmd_text = (" ".join(message[0:4]) + "\n").encode("utf-8")
                            if self.__exec(CommunicationProtocol.START_LEARNING, hash_, cmd_text, int(message[1]), int(message[2])):
                                message = message[4:]
                            else:
                                error = True
                                break
                        else:
                            error = True
                            break
                    else:
                        error = True
                        break

                # Stop learning
                elif message[0] == CommunicationProtocol.STOP_LEARNING:
                    if len(message) > 1:            
                        if message[1].isdigit():
                            hash_ = message[1]
                            cmd_text = (" ".join(message[0:2]) + "\n").encode("utf-8")
                            if self.__exec(CommunicationProtocol.STOP_LEARNING, hash_, cmd_text):
                                message = message[2:]
                            else:
                                error = True
                                break
                        else:
                            error = True
                            break
                    else:
                        error = True
                        break
        
                # Models Ready
                elif message[0] == CommunicationProtocol.MODELS_READY:
                    if len(message) > 1:
                        if message[1].isdigit():
                            if self.__exec(CommunicationProtocol.MODELS_READY, None, None, int(message[1])):
                                message = message[2:]
                            else:
                                error = True
                                break
                        else:
                            error = True
                            break
                    else:
                        error = True
                        break

                # Metrics
                elif message[0] == CommunicationProtocol.METRICS:
                    if len(message) > 5:
                        try:
                            hash_ = message[5]
                            cmd_text = (" ".join(message[0:6]) + "\n").encode("utf-8")
                            if self.__exec(CommunicationProtocol.METRICS, hash_, cmd_text, message[1], int(message[2]), float(message[3]), float(message[4])):
                                message = message[6:]
                            else:
                                error = True
                                break
                        except Exception as e:
                            error = True
                            break
                    else:
                        error = True
                        break

                # Vote train set
                elif message[0] == CommunicationProtocol.VOTE_TRAIN_SET:
                    try:
                        # Divide messages and check length of message
                        close_pos = message.index(CommunicationProtocol.VOTE_TRAIN_SET_CLOSE)
                        node = message[1]
                        vote_msg = message[2:close_pos]
                        hash_ = message[close_pos+1]
                        cmd_text = (" ".join(message[0:close_pos+2]) + "\n").encode("utf-8")

                        if len(vote_msg)%2 != 0:
                            raise Exception("Invalid vote message")
                        message = message[close_pos+2:]

                        # Process vote message
                        votes = []
                        for i in range(0, len(vote_msg), 2):
                            votes.append((vote_msg[i], int(vote_msg[i+1])))

                        if not self.__exec(CommunicationProtocol.VOTE_TRAIN_SET, hash_, cmd_text, node,dict(votes)):
                            error = True
                            break

                    except Exception as e:
                        logging.exception(e)
                        error = True
                        break
                    
                # Models Aggregated
                elif message[0] == CommunicationProtocol.MODELS_AGGREGATED:
                    try:
                        # Divide messages and check length of message
                        close_pos = message.index(CommunicationProtocol.MODELS_AGGREGATED_CLOSE)
                        content = message[1:close_pos]
                        message = message[close_pos+1:]

                        # Get Nodes
                        nodes=[]
                        for n in content:
                            nodes.append(n)

                        # Exec
                        if not self.__exec(CommunicationProtocol.MODELS_AGGREGATED, None, None, nodes):
                            error = True
                            break

                    except Exception as e:
                        logging.exception(e)
                        error = True
                        break

                # Model Initialized
                elif message[0] == CommunicationProtocol.MODEL_INITIALIZED:
                    if self.__exec(CommunicationProtocol.MODEL_INITIALIZED, None, None):
                        message = message[1:]
                    else:
                        error = True
                        break
                    
                # Non Recognized message            
                else:
                    error = True
                    break
                
            # Return
            return self.tmp_exec_msgs,error

    # Exec callbacks
    def __exec(self,action,hash_, cmd_text, *args):
        try:
            # Check if can be executed
            if hash_ not in self.last_messages or hash_ is None:
                self.command_dict[action].execute(*args)
                # Save to gossip
                if hash_ is not None:
                    self.tmp_exec_msgs[hash_] = cmd_text
                    self.add_processed_messages([hash_])
                return True
            return True
        except Exception as e:
            logging.info("Error executing callback: " + str(e))
            logging.exception(e)
            return False


    ########################################
    #    MSG PROCESSING (Static Methods)   #
    ########################################

    def process_connection(message, callback):
        """"
        Static method that checks if the message is a valid connection message and executes the callback (accept connection).

        Args:
            message: The message to check.
            callback: What do if the connection message is legit.
        
        Returns:
            True if connection was accepted, False otherwise.
        """
        message = message.split()
        if len(message) > 4:
            if message[0] == CommunicationProtocol.CONN:
                try:
                    full = message[3] == "1"
                    force = message[4] == "1"
                    callback(message[1], int(message[2]), full, force)
                    return True
                except:
                    return False
            else:
                return False
        else:
            return False

    def check_collapse(msg):
        """"
        Static method that checks if in the message there is a collapse (a binary message (it should fill all the buffer) and a non-binary message before it).

        Actually, collapses can only happen with ``PARAMS`` binary message.

        Args:
            msg: The message to check.

        Returns:
            Length of the collapse (number of bytes to the binary headear). 
        """
        header = CommunicationProtocol.PARAMS.encode("utf-8")
        header_pos = msg.find(header)
        if header_pos != -1 and msg[0:len(header)] != header:
            return header_pos
   
        return 0

    def check_params_incomplete(msg):
        """
        Checks if a params message is incomplete. If the message is complete or is not a params message, it returns 0.
        
        Returns:
            Number of bytes that needs to be complete
        """
        header = CommunicationProtocol.PARAMS.encode("utf-8")
        if msg[0:len(header)] == header:
            if len(msg)<Settings.BLOCK_SIZE:
                return Settings.BLOCK_SIZE-len(msg)

        return 0
    

    #######################################
    #     MSG BUILDERS (Static Methods)   #
    #######################################

    def generate_hased_message(msg):
        """
        Static method that given a non-encoded message generates a hashed and encoded message.

        Args:
            msg: Non encoded message.

        Returns:
            Hashed and encoded message.
        """
        # random number to avoid generating the same hash for a different message (at she same time)
        id = hash(msg+str(datetime.now())+str(random.randint(0,100000)))
        return (msg + " " + str(id) + "\n").encode("utf-8")

    def build_beat_msg(node):
        """ 
        Returns:
            A encoded beat message.
        """
        return CommunicationProtocol.generate_hased_message(CommunicationProtocol.BEAT + " " + node)

    def build_stop_msg():
        """ 
        Returns:
            A encoded stop message.
        """
        return (CommunicationProtocol.STOP + "\n").encode("utf-8")

    def build_connect_to_msg(ip, port):
        """
        Args:
            ip: The ip address to connect to.
            port: The port to connect to.

        Returns:
            A encoded connect to message.
        """
        return (CommunicationProtocol.CONN_TO + " " + ip + " " + str(port) + "\n").encode("utf-8")

    def build_start_learning_msg(rounds, epochs):
        """
        Args:
            rounds: The number of rounds to train.
            epochs: The number of epochs to train.

        Returns:
            A encoded start learning message.
        """
        return CommunicationProtocol.generate_hased_message(CommunicationProtocol.START_LEARNING + " " + str(rounds) + " " + str(epochs))

    def build_stop_learning_msg():
        """
        Returns:
            A encoded stop learning message.
        """
        return CommunicationProtocol.generate_hased_message(CommunicationProtocol.STOP_LEARNING)

    def build_models_ready_msg(round):
        """
        Args:
            round: The last round finished.

        Returns:
            A encoded ready message.
        """
        return (CommunicationProtocol.MODELS_READY + " " + str(round) + "\n").encode("utf-8")

    def build_metrics_msg(node, round, loss, metric):
        """
        Args:
            node: The node that sent the message.
            round: The round when the metrics was calculated.
            loss: The loss of the last round.
            metric: The metric of the last round.
        
        Returns:
            A encoded metrics message.
        """
        return CommunicationProtocol.generate_hased_message(CommunicationProtocol.METRICS + " " + node + " " + str(round) + " " + str(loss) + " " + str(metric))

    def build_vote_train_set_msg(node,votes):
        """
        Args:
            node: The node that sent the message.
            votes: Votes of the node.
        
        Returns:
            A encoded vote train set message.
        """
        aux = ""
        for v in votes:
            aux = aux + " " + v[0]+ " " + str(v[1])
        return CommunicationProtocol.generate_hased_message(CommunicationProtocol.VOTE_TRAIN_SET + " " + node + aux + " " + CommunicationProtocol.VOTE_TRAIN_SET_CLOSE)
        
    def build_models_aggregated_msg(nodes):
        """
        Args:
            nodes: List of strings to indicate aggregated nodes.
        
        Returns:
            A encoded models aggregated message.
        """
        aux = ""
        for n in nodes:
            aux = aux + " " + n
        return (CommunicationProtocol.MODELS_AGGREGATED + aux + " " + CommunicationProtocol.MODELS_AGGREGATED_CLOSE + "\n").encode("utf-8")

    def build_model_initialized_msg():
        """
        Returns:
            A encoded model inicialized message.
        """
        return (CommunicationProtocol.MODEL_INITIALIZED + "\n").encode("utf-8")

    def build_connect_msg(ip, port, broadcast, force):
        """
        Build Handshake message.
        Not Hashed. Special case of message.

        Args:
            ip: The ip address of the node that tries to connect.
            port: The port of the node that tries to connect.
            broadcast: Whether or not to broadcast the message.
            force: Whether or not to force connection.

        Returns:
            A encoded connect message.
        """
        return (CommunicationProtocol.CONN + " " + ip + " " + str(port) + " " + str(broadcast) + " " + str(force) + "\n").encode("utf-8")

    def build_params_msg(data):
        """
        Build model serialized messages.
        Not Hashed. Special case of message (binary message).

        Args:
            data: The model parameters to send (encoded).

        Returns:
            A list of fragments messages of the params.
        """
        # Encoding Headers and ending
        header = CommunicationProtocol.PARAMS.encode("utf-8")
        end = CommunicationProtocol.PARAMS_CLOSE.encode("utf-8")

        # Spliting data
        size = Settings.BLOCK_SIZE - len(header)
        data_msgs = []
        for i in range(0, len(data), size):
            data_msgs.append(header + (data[i:i+size]))

        # Adding closing message
        if len(data_msgs[-1]) + len(end) <= Settings.BLOCK_SIZE:
            data_msgs[-1] += end
            data_msgs[-1] += b'\0' * (Settings.BLOCK_SIZE - len(data_msgs[-1])) # padding to avoid message fragmentation
        else:
            data_msgs.append(header + end)

        return data_msgs


import socket
import threading
import logging
import sys
from p2pfl.communication_protocol import CommunicationProtocol
from p2pfl.encrypter import AESCipher, RSACipher
from p2pfl.gossiper import Gossiper
from p2pfl.settings import Settings
from p2pfl.node_connection import NodeConnection
from p2pfl.heartbeater import Heartbeater
from p2pfl.utils.observer import Events, Observer

class BaseNode(threading.Thread, Observer):
    """
    This class represents a base node in the network (without **FL**). It is a thread, so it's going to process all messages in a background thread using the CommunicationProtocol.

    Args:
        host (str): The host of the node.
        port (int): The port of the node.
        simulation (bool): If False, communication will be encrypted.

    Attributes:
        host (str): The host of the node.
        port (int): The port of the node.
        simulation (bool): If the node is in simulation mode or not. Basically, simulation nodes don't have encryption and metrics aren't sent to network nodes.
        heartbeater (Heartbeater): The heartbeater of the node.
        gossiper (Gossiper): The gossiper of the node.
    """

    #####################
    #     Node Init     #
    #####################

    def __init__(self, host="127.0.0.1", port=None, simulation=True):
        # Node Atributes
        self.host = socket.gethostbyname(host)
        self.port = port
        self.simulation = simulation

        # Super init
        threading.Thread.__init__(self, name = "node-" + self.get_name())
        self._terminate_flag = threading.Event()

        # Setting Up Node Socket (listening)
        self.__node_socket = socket.socket(socket.AF_INET, socket.SOCK_STREAM) #TCP Socket
        if port==None:
            self.__node_socket.bind((host, 0)) # gets a random free port
            self.port = self.__node_socket.getsockname()[1]
        else:
            self.__node_socket.bind((host, port))
        self.__node_socket.listen(50) # no more than 50 connections at queue
        
        # Neighbors
        self.__neighbors = [] # private to avoid concurrency issues
        self.__nei_lock = threading.Lock()

        # Logging
        logging.basicConfig(stream=sys.stdout, level=logging.INFO)

        # Heartbeater and Gossiper
        self.gossiper = None 
        self.heartbeater = None 
        
    def get_addr(self):
        """
        Returns:
            tuple: The address of the node.    
        """
        return self.host,self.port

    def get_name(self):
        """
        Returns:
            str: The name of the node.
        """
        return str(self.get_addr()[0]) + ":" + str(self.get_addr()[1]) 

    #######################
    #   Node Management   #
    #######################
    
    def start(self):
        """
        Starts the node (node loop in a thread). It will listen for new connections and process them. Heartbeater and Gossiper will be started too.

        Note that a node is a thread, so an instance can only be started once.
        """
        # Main Loop
        super().start()
        # Heartbeater and Gossiper
        self.heartbeater = Heartbeater(self.get_name())
        self.gossiper = Gossiper(self.get_name(),self.__neighbors) # thread safe, only read
        self.heartbeater.add_observer(self)
        self.gossiper.add_observer(self)
        self.heartbeater.start()
        self.gossiper.start()
    
    def stop(self): 
        """
        Stops the node. Heartbeater and Gossiper will be stopped too.
        """
        self._terminate_flag.set()
        try:
            # Send a self message to the loop to avoid the wait of the next recv
            self.__send(self.host,self.port,b"")
        except:
            pass

    ########################
    #   Main Thread Loop   #
    ########################

    def run(self):
        """
        Main loop of the node, when a node is running, this method is being executed. It will listen for new connections and process them.
        """
        # Process new connections loop
        logging.info('({}) Node started'.format(self.get_name()))
        while not self._terminate_flag.is_set(): 
            try:
                (ns, _) = self.__node_socket.accept()
                msg = ns.recv(Settings.BLOCK_SIZE)
                
                # Process new connection
                if msg:
                    msg = msg.decode("UTF-8")
                    callback = lambda h,p,fu,fc: self.__process_new_connection(ns, h, p, fu, fc)
                    if not CommunicationProtocol.process_connection(msg,callback):
                        ns.close()
            except Exception as e:
                logging.exception(e)
       
        # Stop Heartbeater and Gossiper
        self.heartbeater.stop()
        self.gossiper.stop()

        # Stop Node
        logging.info('({}) Stopping node. Disconnecting from {} nodos'.format(self.get_name(), len(self.__neighbors))) 
        nei_copy_list = self.get_neighbors()
        for n in nei_copy_list:
            n.stop()
        self.__node_socket.close()


    def __process_new_connection(self, node_socket, h, p, full, force):
        try:
            # Check if connection with the node already exist
            self.__nei_lock.acquire()
            if self.get_neighbor(h,p,thread_safe=False) == None:

                # Check if ip and port are correct
                s = socket.socket(socket.AF_INET, socket.SOCK_STREAM) 
                s.settimeout(2)
                result = s.connect_ex((h,p)) 
                s.close()

                # Encryption
                aes_cipher = None
                if not self.simulation:
                    # Asymmetric
                    rsa = RSACipher()
                    node_socket.sendall(rsa.get_key())
                    rsa.load_pair_public_key(node_socket.recv(len(rsa.get_key())))

                    # Symmetric
                    aes_cipher = AESCipher()
                    node_socket.sendall(aes_cipher.get_key())

                # Add neighboor
                if result == 0:
                    logging.info('{} Connection accepted with {}:{}'.format(self.get_name(),h,p))
                    nc = NodeConnection(self.get_name(),node_socket,(h,p),aes_cipher)
                    nc.add_observer(self)
                    self.__neighbors.append(nc)
                    nc.start(force=force)
                    
                    if full:
                        self.broadcast(CommunicationProtocol.build_connect_to_msg(h,p),exc=[nc],thread_safe=False)
            else:
                node_socket.close()
            
            self.__nei_lock.release()
            
        except Exception as e:
            logging.debug('({}) Connection refused with {}:{}'.format(self.get_name(),h,p))
            self.__nei_lock.release()
            node_socket.close()
            self.rm_neighbor(nc)

    #############################
    #  Neighborhood management  #
    #############################

    # Create a tcp socket and send data
    def __send(self, h, p, data, persist=False): 
        s = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        s.connect((h, p))
        s.sendall(data)
        if persist:
            return s
        else:
            s.close()
            return None

    def connect_to(self, h, p, full=False, force=False):
        """"
        Connects a node to another.
        
        Args:
            h (str): The host of the node.
            p (int): The port of the node.
            full (bool): If True, the node will be connected to the entire network.
            force (bool): If True, the node will be connected even though it should not be.

        Returns:
            node: The node that has been connected to.
        """
        if full:
            full = "1"
        else:
            full = "0"
        
        if force:
            force = "1"
        else:
            force = "0"

        try:
            # Check if connection with the node already exist
            h = socket.gethostbyname(h)
            self.__nei_lock.acquire()
            if self.get_neighbor(h,p,thread_safe=False) == None:
            
                # Send connection request
                msg=CommunicationProtocol.build_connect_msg(self.host,self.port,full,force)
                s = self.__send(h,p,msg,persist=True)

                # Encryption
                aes_cipher = None
                if not self.simulation:
                    # Asymetric
                    rsa = RSACipher()
                    rsa.load_pair_public_key(s.recv(len(rsa.get_key())))
                    s.sendall(rsa.get_key())
                    # Symetric
                    aes_cipher = AESCipher(key=s.recv(AESCipher.key_len()))

                # Add socket to neighbors
                logging.info("{} Connected to {}:{}".format(self.get_name(),h,p))
                nc = NodeConnection(self.get_name(),s,(h,p),aes_cipher)
                nc.add_observer(self)
                self.__neighbors.append(nc)
                nc.start(force=force)
                self.__nei_lock.release()
                return nc
        
            else:
                logging.info("{} Already connected to {}:{}".format(self.get_name(),h,p))
                self.__nei_lock.release()
                return None
    
        except Exception as e:
            logging.info("{} Can't connect to the node {}:{}".format(self.get_name(),h,p))
            logging.exception(e)
            try:
                self.__nei_lock.release()
            except:
                pass
            return None

    def disconnect_from(self, h, p):
        """
        Disconnects from a node.
        
        Args:
            h (str): The host of the node.
            p (int): The port of the node.
        """
        self.get_neighbor(h,p).stop()
            
    def get_neighbor(self, h, p, thread_safe=True):
        """
        Get a ``NodeConnection`` from the neighbors list.

        Args:
            h (str): The host of the node.
            p (int): The port of the node.

        Returns:
            NodeConnection: The connection with the node.
        """
        if thread_safe:
            self.__nei_lock.acquire()

        return_node = None    
        for n in self.__neighbors:
            if n.get_addr() == (h,p):
                return_node = n
                break
            
        if thread_safe:
            self.__nei_lock.release()
        
        return return_node

    def get_neighbors(self):
        """
        Returns:
            list: The neighbors of the node.
        """
        self.__nei_lock.acquire()
        n = self.__neighbors.copy()
        self.__nei_lock.release()
        return n

    def rm_neighbor(self,n):
        """
        Removes a neighboor from the neighbors list.

        Args:
            n (NodeConnection): The neighboor to be removed.
        """
        self.__nei_lock.acquire()
        try:
            self.__neighbors.remove(n)
            n.stop()
        except:
            pass
        self.__nei_lock.release()

    def get_network_nodes(self):
        """
        Returns:
            list: The nodes of the network.
        """
        return self.heartbeater.get_nodes()
      
    ##########################
    #     Msg management     #
    ##########################

    def broadcast(self, msg, exc=[], thread_safe=True):
        """
        Broadcasts a message to all the neighbors.

        Args:
            msg (str): The message to be broadcasted.
            exc (list): The neighbors to be excluded.
            thread_safe (bool): If True, the broadcast will access the neighbors list in a thread safe mode.

        """
        if thread_safe:
            self.__nei_lock.acquire()

        for n in self.__neighbors:
            if not (n in exc):
                n.send(msg)

        if thread_safe:
            self.__nei_lock.release()
        
    ###########################
    #     Observer Events     #
    ###########################

    def update(self,event,obj):
        """
        Observer update method. Used to handle events that can occur with the different components and connections of the node.
        
        Args:
            event (Events): Event that has occurred.
            obj: Information about the change or event.
        """
        if event == Events.END_CONNECTION_EVENT:
            self.rm_neighbor(obj)

        elif event == Events.NODE_CONNECTED_EVENT:
            n,_ = obj
            n.send(CommunicationProtocol.build_beat_msg(self.get_name())) 

        elif event == Events.CONN_TO_EVENT:
            self.connect_to(obj[0], obj[1], full=False)

        elif event == Events.SEND_BEAT_EVENT:
            self.broadcast(CommunicationProtocol.build_beat_msg(self.get_name())) 

        elif event == Events.GOSSIP_BROADCAST_EVENT:
            self.broadcast(obj[0],exc=obj[1]) 
            
        elif event == Events.PROCESSED_MESSAGES_EVENT:
            node, msgs = obj
            # Comunicate to connections the new messages processed
            for nc in self.__neighbors:
                if nc != node:
                    nc.add_processed_messages(list(msgs.keys()))
            # Gossip the new messages
            self.gossiper.add_messages(list(msgs.values()),node)

        elif event == Events.BEAT_RECEIVED_EVENT:
            self.heartbeater.add_node(obj)




from torch.utils.data import DataLoader, Subset, random_split
from torchvision import transforms
from torchvision.datasets import MNIST
from math import floor
from pytorch_lightning import LightningDataModule
# To Avoid Crashes with a lot of nodes
import torch.multiprocessing
torch.multiprocessing.set_sharing_strategy('file_system')

#######################################
#    FederatedDataModule for MNIST    #
#######################################

class MnistFederatedDM(LightningDataModule):
    """
    LightningDataModule of partitioned MNIST. Its used to generate **IID** distribucions over MNIS. Toy Problem.

    Args:
        sub_id: Subset id of partition. (0 <= sub_id < number_sub)
        number_sub: Number of subsets.
        batch_size: The batch size of the data.
        num_workers: The number of workers of the data.
        val_percent: The percentage of the validation set.
    """

    # Singleton
    mnist_train = None
    mnist_val = None

    def __init__(self, sub_id=0, number_sub=1, batch_size=32, num_workers=4, val_percent=0.1, iid=True):
        super().__init__()
        self.sub_id=sub_id
        self.number_sub=number_sub
        self.batch_size=batch_size
        self.num_workers=num_workers
        self.val_percent=val_percent
        
        # Singletons of MNIST train and test datasets
        if MnistFederatedDM.mnist_train is None:
            MnistFederatedDM.mnist_train = MNIST("", train=True, download=True, transform=transforms.ToTensor())
            if not iid:
                sorted_indexes = MnistFederatedDM.mnist_train.targets.sort()[1]
                MnistFederatedDM.mnist_train.targets = MnistFederatedDM.mnist_train.targets[sorted_indexes]
                MnistFederatedDM.mnist_train.data = MnistFederatedDM.mnist_train.data[sorted_indexes]
        if MnistFederatedDM.mnist_val is None:
            MnistFederatedDM.mnist_val = MNIST("", train=False, download=True, transform=transforms.ToTensor())
            if not iid:
                sorted_indexes = MnistFederatedDM.mnist_val.targets.sort()[1]
                MnistFederatedDM.mnist_val.targets = MnistFederatedDM.mnist_val.targets[sorted_indexes]
                MnistFederatedDM.mnist_val.data = MnistFederatedDM.mnist_val.data[sorted_indexes]
        if self.sub_id+1 > self.number_sub:
            raise("Not exist the subset {}".format(self.sub_id))

        # Training / validation set
        trainset = MnistFederatedDM.mnist_train
        rows_by_sub = floor(len(trainset)/self.number_sub)
        tr_subset = Subset(trainset,range(self.sub_id*rows_by_sub,(self.sub_id+1)*rows_by_sub))
        mnist_train, mnist_val = random_split(tr_subset, [round(len(tr_subset)*(1-self.val_percent)), round(len(tr_subset)*self.val_percent)])
        
        # Test set
        testset = MnistFederatedDM.mnist_val
        rows_by_sub = floor(len(testset)/self.number_sub)
        te_subset = Subset(testset,range(self.sub_id*rows_by_sub,(self.sub_id+1)*rows_by_sub))

        if len(testset) < self.number_sub:
            raise("Too much partitions")
        
        #DataLoaders
        self.train_loader = DataLoader(mnist_train, batch_size=self.batch_size, shuffle=True, num_workers=self.num_workers)
        self.val_loader   = DataLoader(mnist_val, batch_size=self.batch_size, shuffle=False, num_workers=self.num_workers)
        self.test_loader  = DataLoader(te_subset, batch_size=self.batch_size, shuffle=False, num_workers=self.num_workers)
        print("Train: {} Val:{} Test:{}".format(len(mnist_train),len(mnist_val),len(te_subset)))

    
    def train_dataloader(self):
        """
        """
        return self.train_loader

    def val_dataloader(self):
        """
        """
        return self.val_loader

    def test_dataloader(self):
        """
        """
        return self.test_loader


import time
import threading
from p2pfl.settings import Settings
from p2pfl.utils.observer import Events, Observable

##################
#    Gossiper    #
##################

class Gossiper(threading.Thread, Observable):
    """
    Thread based gossiper. It gossip messages from list of pending messages. ``Settings.GOSSIP_MESSAGES_PER_ROUND`` are sended per iteration (`Settings.GOSSIP_FREC` times per second).

    Communicates with node via observer pattern.
    
    Args:
        nodo_padre (str): Name of the parent node.
        neighbors (list): List of neighbors.

    """
    def __init__(self, node_name, neighbors):
        Observable.__init__(self)
        threading.Thread.__init__(self, name=("gossiper-" + node_name ))
        self.__neighbors = neighbors # list as reference of the original neighbors list
        self.__msgs = {}
        self.__add_lock = threading.Lock()
        self.__terminate_flag = threading.Event()

    def add_messages(self, msgs, node):
        """
        Add messages to the list of pending messages.

        Args:
            msgs (list): List of messages to add.
            node (Node): Node that sent the messages.
        """
        self.__add_lock.acquire()
        for msg in msgs:
            self.__msgs[msg] = [node]
        self.__add_lock.release()

    def run(self):
        """
        Gossiper Main Loop. Sends `Settings.GOSSIP_MODEL_SENDS_BY_ROUND` messages `Settings.GOSSIP_FREC` times per second.        
        """
        while not self.__terminate_flag.is_set():

            messages_left = Settings.GOSSIP_MESSAGES_PER_ROUND
            
            # Lock
            self.__add_lock.acquire()
            begin = time.time()
            
            # Send to all the nodes except the ones that the message was already sent to
            if len(self.__msgs) > 0:
                msg_list = list(self.__msgs.items()).copy()
                nei = set(self.__neighbors.copy()) # copy to avoid concurrent problems

                for msg,nodes in msg_list:
                    nodes = set(nodes)
                    sended = len(nei - nodes)

                    if messages_left - sended >= 0:
                        self.notify(Events.GOSSIP_BROADCAST_EVENT, (msg,list(nodes)))
                        del self.__msgs[msg]
                        messages_left = messages_left - sended
                        if messages_left == 0:
                            break
                    else:
                        # Lists to concatenate / Sets to difference
                        excluded = (list(nei - nodes))[:abs(messages_left - sended)]
                        self.notify(Events.GOSSIP_BROADCAST_EVENT, (msg,list(nodes)+excluded))
                        self.__msgs[msg] = list(nodes) + list(nei - set(excluded))
                        break

            # Unlock
            self.__add_lock.release()
            
            # Wait to guarantee the frequency of gossipping
            time_diff = time.time() - begin
            time_sleep = 1/Settings.GOSSIP_MESSAGES_FREC-time_diff
            if time_sleep > 0:
                time.sleep(time_sleep)

    def stop(self):
        """
        Stop the gossiper.
        """
        self.__terminate_flag.set()


"""
Module to define constants for the p2pfl system.
"""

import logging
from p2pfl.encrypter import AESCipher

###################
# Global Settings #
###################

class Settings():
    """
    Class to define global settings for the p2pfl system.
    """

    # TCP Block Size
    BLOCK_SIZE = 2048
    """
    Max ammount of bytes to read from the buffer at a time. If ```BLOCK_SIZE`` is not divisible by the block size used for symetric encryption it will be rounded to the next closest value.

    Try to strike a balance between hyper-segmentation and excessively large block size.
    """
    rest = BLOCK_SIZE % AESCipher.get_block_size()
    if rest != 0:
        new_value = BLOCK_SIZE + AESCipher.get_block_size() - rest
        logging.info("Changing buffer size to %d. %d is incompatible with the AES block size.", BLOCK_SIZE, new_value)
        BLOCK_SIZE = new_value

    # TIMEOUT'S
    NODE_TIMEOUT = 20
    """
    Timeout (seconds) for a node to be considered dead.
    """
    VOTE_TIMEOUT = 60
    """
    Timeout (seconds) for a node to wait for a vote.
    """
    AGGREGATION_TIMEOUT = 60
    """
    Timeout (seconds) for a node to wait for other models. Timeout stars when the first model is added.
    """

    # HEARTBEATER
    HEARTBEAT_PERIOD = 4
    """
    Period (seconds) for the node to send a heartbeat.
    """
    HEARTBEATER_REFRESH_NEIGHBORS_BY_PERIOD = 4
    """
    Times by period to refresh the neighbors list.
    """
    WAIT_HEARTBEATS_CONVERGENCE = 10
    """
    Time (seconds) to wait for the heartbeats to converge.
    """
    
    # TRAIN SET 
    TRAIN_SET_SIZE = 10
    """
    Size of the training set.
    """
    TRAIN_SET_CONNECT_TIMEOUT = 5
    """
    Timeout (seconds) to wait for a node to connect to the training set.
    """

    # GOSSIP Messages
    AMOUNT_LAST_MESSAGES_SAVED = 100 
    """
    Amount of last messages saved. Used to avoid gossiping cycles.
    """
    GOSSIP_MESSAGES_FREC = 100
    """
    Frequency of gossiping messages (times by second).
    """
    GOSSIP_MESSAGES_PER_ROUND = 100
    """
    Amount of gossip messages per round.
    """

    # GOSSIP Models
    GOSSIP_EXIT_ON_X_EQUAL_ROUNDS = 20
    """
    Amount of equal rounds to exit gossiping. Careful, a low value can cause an early stop of gossiping.
    """
    GOSSIP_MODELS_FREC = 1
    """
    Frequency of gossiping models (times by second).
    """
    GOSSIP_MODELS_PER_ROUND = 2
    """
    Amount of gossip models per round.
    """



################################
#    NodeLearning Interface    #  -->  Template Pattern
################################

class NodeLearner:
    """
    Template to implement learning processes, iincluding metric monitoring during training.
    """

    def set_model(self, model): 
        """
        Set the model of the learner. (not wheights)

        Args:
            model: The model of the learner.

        Raises:
            ModelNotMatchingError: If the model is not matching the learner.
        """
        pass

    def set_data(self, data): 
        """
        Set the data of the learner. It is used to fit the model.

        Args:
            data: The data of the learner. 
        """
        pass

    def encode_parameters(self, params=None, contributors=None, weight=None):
        """
        Encode the parameters of the model. (binary)
        If params are not provided, self parameters are encoded.
        
        Args:
            params: The parameters of the model. (non-binary)
            contributors: The contributors of the model.
            weight: The weight of the model. 
            
        Returns:
            The encoded parameters of the model (params, contributors, weight).
        """
        pass

    def decode_parameters(self,data): 
        """
        Decode the parameters of the model. (binary)
        
        Args:
            data: The encoded parameters of the model.

        Returns:
            The decoded parameters of the model. (params, contributors, weight)

        Raises:
            DecodingParamsError: If the decoding of the parameters fails.
        """
        pass

    def check_parameters(self, params): 
        """
        Check if the parameters are compatible with the model.

        Args:
            params: The parameters to check. (non-binary)

        Returns:
            True if the parameters are compatible with the model.
        """
        pass
    
    def set_parameters(self, params): 
        """
        Set the parameters of the model.

        Args:
            params: The parameters of the model. (non-binary)

        Raises:
            ModelNotMatchingError: If the model is not matching the learner.
        """
        pass

    def get_parameters(self): 
        """
        Get the parameters of the model.

        Returns:
            The parameters of the model. (non-binary)
        """
        pass

    def set_epochs(self, epochs): 
        """
        Set the number of epochs of the model.

        Args:
            epochs: The number of epochs of the model.
        """
        pass

    def fit(self): 
        """
        Fit the model.
        """
        pass

    def interrupt_fit(self): 
        """
        Interrupt the fit process.
        """
        pass

    def evaluate(self): 
        """
        Evaluate the model. With a given parameters.
        """
        pass

    def log_validation_metrics(self, loss, metric, round=None, name=None):
        """
        Log the validation metrics. It also can be used to log the othe node metrics.
        """
        pass

    def get_num_samples(self): 
        """
        Get the number of samples of the model.

        Returns:
            The number of samples of the model.
        """
        pass

    def init(self): 
        """
        Init the learner.
        """
        pass

    def close(self): 
        """
        Close the learner.
        """
        pass

    def finalize_round(self):
        """
        Determine the end of the round.
        """
        pass



import torch
from torch import nn
from torch.nn import functional as F
import pytorch_lightning as pl
from torchmetrics import Accuracy

###############################
#    Multilayer Perceptron    #
###############################

IMAGE_SIZE = 28

class CNN(pl.LightningModule):
    """
    Convolutional Neural Network (CNN) to solve MNIST with PyTorch Lightning.
    """

    def __init__(self, in_channels=1, out_channels=10, metric=Accuracy(), lr_rate=0.001, seed=None):

        # Set seed for reproducibility iniciialization
        if seed is not None:
            torch.manual_seed(seed)
            torch.cuda.manual_seed_all(seed)

        super().__init__()
        self.metric = metric
        self.lr_rate = lr_rate

        self.conv1 = nn.Conv2d(in_channels=in_channels, out_channels=32, kernel_size=(5, 5), padding="same")
        self.relu = nn.ReLU()
        self.pool1 = nn.MaxPool2d(kernel_size=(2,2), stride=2)
        self.conv2 = nn.Conv2d(in_channels=32, out_channels=64, kernel_size=(5, 5), padding="same")
        self.pool2 = nn.MaxPool2d(kernel_size=(2,2), stride=2)
        self.l1 = nn.Linear(7 * 7 * 64, 2048)
        self.l2 = nn.Linear(2048, out_channels)

        self.loss_fn = nn.CrossEntropyLoss()

    def forward(self, x):
        """
        """
        input_layer = x.view(-1, 1, IMAGE_SIZE, IMAGE_SIZE)
        conv1 = self.relu(self.conv1(input_layer))
        pool1 = self.pool1(conv1)
        conv2 = self.relu(self.conv2(pool1))
        pool2 = self.pool2(conv2)
        pool2_flat = pool2.reshape(-1, 7 * 7 * 64)

        dense = self.relu(self.l1(pool2_flat))
        logits = self.l2(dense)

        return logits

    def configure_optimizers(self):
        """
        """
        return torch.optim.Adam(self.parameters(), lr=self.lr_rate)

    def training_step(self, batch, batch_id):
        """
        """
        x, y = batch
        loss = self.loss_fn(self(x), y)
        self.log("train_loss", loss, prog_bar=True)
        return loss 


    def validation_step(self, batch, batch_idx):
        """
        """
        x, y = batch
        logits = self(x)
        loss = self.loss_fn(self(x), y)
        out = torch.argmax(logits, dim=1)
        metric = self.metric(out, y) 
        self.log("val_loss", loss, prog_bar=True)
        self.log("val_metric", metric, prog_bar=True)    
        return loss

    def test_step(self, batch, batch_idx):
        """
        """
        x, y = batch
        logits = self(x)
        loss = self.loss_fn(self(x), y)
        out = torch.argmax(logits, dim=1)
        metric = self.metric(out, y) 
        self.log("test_loss", loss, prog_bar=True)
        self.log("test_metric", metric, prog_bar=True)
        return loss
        


import threading
import logging
from p2pfl.settings import Settings
from p2pfl.utils.observer import Events, Observable
   
class Aggregator(threading.Thread, Observable):
    """
    Class to manage the aggregation of models. Its a thread so, aggregation will be done in background if all models was added or timeouts have gone. 
    Also its a observable so, it will notify the node when the aggregation was done.

    Args:
        node_name: (str): String with the name of the node.
    """
    def __init__(self, node_name="unknown"):
        self.node_name = node_name
        threading.Thread.__init__(self, name = "aggregator-" + node_name)
        self.daemon = True
        Observable.__init__(self)
        self.__train_set = []
        self.__waiting_aggregated_model = False
        self.__aggregated_waited_model = False
        self.__models = {}
        self.__lock = threading.Lock()
        self.__aggregation_lock = threading.Lock()
        self.__aggregation_lock.acquire()
        self.__thread_executed = False

    def run(self):
        """
        Wait for the aggregation to be done or timeout. Then aggregate the models and notify.
        """
        self.__thread_executed = True

        # Wait for all models to be added or TIMEOUT
        self.__aggregation_lock.acquire(timeout=Settings.AGGREGATION_TIMEOUT) 
        
        # Check if node still running (could happen if aggregation thread was a residual thread)
        if self.__train_set == []:
            logging.info("({}) Shutting Down Aggregator Process".format(self.node_name))
            self.notify(Events.AGGREGATION_FINISHED_EVENT,None) # To avoid residual trainning-thread
            return
        
        # Start aggregation
        n_model_aggregated = sum([len(nodes.split()) for nodes in list(self.__models.keys())])
        if n_model_aggregated != len(self.__train_set):
            logging.info("({}) Aggregating models, timeout reached. Missing models: {}".format(self.node_name,set(self.__train_set)-set(self.__models.keys())))
        else:
            logging.info("({}) Aggregating models.".format(self.node_name))

        # Notify node
        self.notify(Events.AGGREGATION_FINISHED_EVENT,self.aggregate(self.__models)) 

    def aggregate(self,models): 
        """
        Aggregate the models.
        """
        print("Not implemented")
            
    def set_nodes_to_aggregate(self, l):
        """
        List with the name of nodes to aggregate.

        Args:
            l: List of nodes to aggregate. Empty for no aggregation.
        """
        self.__train_set = l

    def set_waiting_aggregated_model(self):
        """
        Indicates that the node is waiting for an aggregation. It won't participate in aggregation process.
        """
        self.__waiting_aggregated_model = True

    def add_model(self, model, nodes, weight):
        """
        Add a model. The first model to be added starts the `run` method (timeout).

        Args:
            model: Model to add.
            nodes: Nodes that colaborated to get the model.
            weight: Number of samples used to get the model.
        """
        if self.__waiting_aggregated_model and not self.__aggregated_waited_model:
            logging.info("({}) Recived an aggregated model.".format(self.node_name))
            self.__aggregated_waited_model = True
            self.notify(Events.AGGREGATION_FINISHED_EVENT,model) 
        else:
            if nodes is not None:
                self.__lock.acquire()

                # Start aggregation timeout
                if self.__train_set != [] and not self.__thread_executed:
                    self.start()  

                # Get a list of nodes added
                models_added = [n.split() for n in list(self.__models.keys())] 
                models_added = [element for sublist in models_added for element in sublist] # Flatten list
                
                # Check if aggregation is needed
                if len(self.__train_set)>len(models_added):
                    # Check if all nodes are in the train_set 
                    if all([n in self.__train_set for n in nodes]): 
                        # Check if all nodes are not aggregated                    
                        if all([n not in models_added for n in nodes]):
                            # Aggregate model
                            self.__models[" ".join(nodes)] = ((model, weight))
                            logging.info("({}) Model added ({}/{}) from {}".format(self.node_name, str(len(models_added)+len(nodes)), str(len(self.__train_set)), str(nodes)))
                            # Check if all models have been added
                            self.check_and_run_aggregation()
                            # Build response 
                            response = models_added + nodes
                            # Unloock
                            self.__lock.release()
                            
                            return response
                        else:
                            self.__lock.release()
                            logging.debug("({}) Can't add a model that has already been added {}".format(self.node_name, nodes))
                    else:
                        self.__lock.release()
                        logging.debug("({}) Can't add a model from a node ({}) that is not in the training test.".format(self.node_name, nodes))                
                else:
                    self.__lock.release()

        return None
                
    def get_partial_aggregation(self,except_nodes):
        """
        Get the partial aggregation of the models.

        Args:
            except_nodes: Nodes to exclude.

        Returns:
            (model, nodes, weight): Model, nodes and number of samples for the partial aggregation.
        """
        dict_aux = {}
        nodes_aggregated = []
        aggregation_weight = 0
        models = self.__models.copy()
        for n,(m,s) in list(models.items()):
            splited_nodes = n.split() 
            if all([n not in except_nodes for n in splited_nodes]):
                dict_aux[n] = (m,s)
                nodes_aggregated += splited_nodes
                aggregation_weight += s
        
        # If there are no models to aggregate
        if len(dict_aux) == 0:
            return None,None,None

        return (self.aggregate(dict_aux), nodes_aggregated, aggregation_weight)
            
    def check_and_run_aggregation(self,force=False):
        """
        Check if all models have been added and start aggregation if so.

        Args:
            force: If true, aggregation will be started even if not all models have been added.
        """
        models_added = [nodes.split() for nodes in list(self.__models.keys())] 
        models_added = [element for sublist in models_added for element in sublist] # Flatten list
        # Try Unloock
        try:
            if (force or len(models_added)>=len(self.__train_set)) and self.__train_set!=[]: 
                self.__aggregation_lock.release()
        except:
            pass

    def clear(self):
        """
        Clear all for a new aggregation.
        """
        observers = self.get_observers()
        self.__init__(node_name=self.node_name)
        for o in observers:
            self.add_observer(o)



import socket
import threading
import logging
from p2pfl.command import *
from p2pfl.communication_protocol import CommunicationProtocol
from p2pfl.settings import Settings
from p2pfl.utils.observer import Events, Observable

########################
#    NodeConnection    #
########################

class NodeConnection(threading.Thread, Observable):
    """
    This class represents a connection to a node. It is a thread, so it's going to process all messages in a background thread using the CommunicationProtocol.

    The NodeConnection can recive many messages in a single recv and exists 2 kinds of messages:
        - Binary messages (models)
        - Text messages (commands)

    Carefully, if the connection is broken, it will be closed. If the user wants to reconnect, he/she should create a new connection.

    Args:
        parent_node: The parent node of this connection.
        s: The socket of the connection.
        addr: The address of the node that is connected to.
    """

    ############## 
    #    Init    # 
    ##############

    def __init__(self, parent_node_name, s, addr, aes_cipher, tcp_buffer_size=(None,None)):
        # Init supers
        threading.Thread.__init__(self, name = ("node_connection-" + parent_node_name + "-" + str(addr[0]) + ":" + str(addr[1])))
        Observable.__init__(self)
        # Connection Loop
        self.__terminate_flag = threading.Event()
        self.__socket = s
        self.__socket_lock = threading.Lock()

        if tcp_buffer_size[0] is not None:
            self.__socket.setsockopt(socket.SOL_SOCKET, socket.SO_RCVBUF, tcp_buffer_size[0])

        if tcp_buffer_size[1] is not None:
            self.__socket.setsockopt(socket.SOL_SOCKET, socket.SO_SNDBUF, tcp_buffer_size[1])

        # Atributes
        self.__addr = addr
        self.__param_bufffer = b""
        self.__model_ready = -1
        self.__aes_cipher = aes_cipher
        self.__model_initialized = False
        self.__models_aggregated = []
        # Communication Protocol
        self.comm_protocol = CommunicationProtocol({
            CommunicationProtocol.BEAT: Beat_cmd(self),
            CommunicationProtocol.STOP: Stop_cmd(self),
            CommunicationProtocol.CONN_TO: Conn_to_cmd(self),
            CommunicationProtocol.START_LEARNING: Start_learning_cmd(self),
            CommunicationProtocol.STOP_LEARNING: Stop_learning_cmd(self),
            CommunicationProtocol.PARAMS: Params_cmd(self),
            CommunicationProtocol.MODELS_READY: Models_Ready_cmd(self),
            CommunicationProtocol.METRICS: Metrics_cmd(self),
            CommunicationProtocol.VOTE_TRAIN_SET: Vote_train_set_cmd(self),
            CommunicationProtocol.MODELS_AGGREGATED: Models_aggregated_cmd(self),
            CommunicationProtocol.MODEL_INITIALIZED: Model_initialized_cmd(self),
        })

    ############## 
    #    Name    # 
    ##############

    def get_addr(self):
        """
        Returns:
            (ip, port) The address of the node that is connected to. 
        """
        return self.__addr

    def get_name(self):
        """
        Returns:
            The name of the node connected to.
        """
        return self.__addr[0] + ":" + str(self.__addr[1])

    ################### 
    #    Main Loop    # 
    ###################

    def start(self, force=False):
        """
        Start the connection. It will start the connection thread, this thread is receiving messages and processing them.
        
        Args:
            force: Determine if connection is going to keep alive even if it should not.
        """
        self.notify(Events.NODE_CONNECTED_EVENT, (self,force))
        return super().start()

    def run(self):
        """
        NodeConnection loop. Recive and process messages.
        """
        self.__socket.settimeout(Settings.NODE_TIMEOUT)
        amount_pending_params = 0
        param_buffer = b""
        while not self.__terminate_flag.is_set():
            try:
                # Recive message
                og_msg = b""
                if amount_pending_params == 0:
                    og_msg = self.__socket.recv(Settings.BLOCK_SIZE)

                else:
                    pending_fragment = self.__socket.recv(amount_pending_params)
                    og_msg = param_buffer + pending_fragment #alinear el colapso
                    param_buffer = b""
                    amount_pending_params = 0

                # Decrypt message
                if self.__aes_cipher is not None:
                    # Guarantee block size (if TCP sctream is slow)
                    bytes_to_block_size = len(og_msg)%self.__aes_cipher.bs
                    # Decrypt
                    if bytes_to_block_size != 0:
                        msg = self.__aes_cipher.decrypt(og_msg + self.__socket.recv(bytes_to_block_size))
                    else:
                        msg = self.__aes_cipher.decrypt(og_msg)
                else:
                    msg = og_msg
            
                # Process messages
                if msg!=b"":
                    # Check if fragments are incomplete (collapse / TCP stream slow)
                    overflow = CommunicationProtocol.check_collapse(msg)
                    if overflow>0:
                        param_buffer = og_msg[overflow:]
                        amount_pending_params = Settings.BLOCK_SIZE-len(param_buffer)
                        msg = msg[:overflow]
                        logging.debug("({}) (NodeConnection Run) Collapse detected: {}".format(self.get_name(), msg))

                    else:
                        # Check if all bytes of param_buffer are received
                        amount_pending_params = CommunicationProtocol.check_params_incomplete(msg) 
                        if amount_pending_params != 0:
                            param_buffer = msg
                            continue

                    # Process message
                    exec_msgs,error = self.comm_protocol.process_message(msg)
                    if len(exec_msgs) > 0:
                        self.notify(Events.PROCESSED_MESSAGES_EVENT, (self,exec_msgs)) # Notify the parent node

                    # Error happened
                    if error:
                        self.__terminate_flag.set()
                        logging.info("({}) An error happened. Last error: {}".format(self.get_name(),msg))       

            except socket.timeout:
                logging.info("({}) (NodeConnection Loop) Timeout".format(self.get_name()))
                self.__terminate_flag.set()
                break

            except Exception as e:
                logging.info("({}) (NodeConnection Loop) Exception: ".format(self.get_name()) + str(e))
                self.__terminate_flag.set()
                break
        
        #Down Connection
        logging.info("Closed connection: {}".format(self.get_name()))
        self.notify(Events.END_CONNECTION_EVENT, self) 
        self.__socket.close()

    def stop(self,local=False):
        """
        Stop the connection. Stops the main loop and closes the socket.
        
        Args:
            local: If true, the connection will be closed without notifying the other node.
        """
        if not local:
            self.send(CommunicationProtocol.build_stop_msg())
        self.__terminate_flag.set()

    ############################
    #    Processed Messages    #
    ############################

    def add_processed_messages(self,msgs):
        """
        Add to a list of communication protocol messages that have been processed. (By other nodes)
        Messages are added to avoid multiple processing of the same messages, also the non-processing
        of these messges avoid cycles in the network.
 
        Args:
            msgs: The list of messages that have been processed.

        """
        self.comm_protocol.add_processed_messages(msgs)

    ############################
    #    Model Ready Rounds    #
    ############################

    def set_model_ready_status(self,round):
        """
        Set the last ready round of the other node.

        Args:
            round: The last ready round of the other node.
        """
        self.__model_ready = round

    def get_model_ready_status(self):
        """
        Returns:
            The last ready round of the other node.
        """
        return self.__model_ready
    
    ###########################
    #    Model Initialized    #
    ###########################

    def set_model_initialized(self, value):
        """
        Set the model initialized.

        Args:
            value: True if the model is initialized, false otherwise.
        """
        self.__model_initialized = value

    def get_model_initialized(self):
        """
        Returns:
            The model initialized.
        """
        return self.__model_initialized

    ##########################
    #    Models Aggregated    #
    ##########################

    def add_models_aggregated(self,models):
        """
        Add the models aggregated.
        
        Args:
            models: Models aggregated.
        """
        self.__models_aggregated = list(set(models+self.__models_aggregated))

    def clear_models_aggregated(self):
        """
        Clear models aggregated.
        """
        self.__models_aggregated = []
    def get_models_aggregated(self):
        """
        Returns:
            The models aggregated.
        """
        return self.__models_aggregated
    
    #######################
    #    Params Buffer    #
    #######################

    def add_param_segment(self,data):
        """
        Add a segment of parameters to the buffer.
        
        Args:
            data: The segment of parameters.
        """
        self.__param_bufffer = self.__param_bufffer + data

    def get_params(self):
        """
        Returns:
            The parameters buffer content.
        """
        return self.__param_bufffer

    def clear_buffer(self):
        """
        Clear the params buffer.
        """
        self.__param_bufffer = b""

    ##################
    #    Messages    # 
    ##################

    def send(self, data): 
        """
        Tries to send a message to the other node.

        Args:
            data: The message to send.

        Returns:
            True if the message was sent, False otherwise.

        """    
        # Check if the connection is still alive
        if not self.__terminate_flag.is_set():
            try:
                # Encrypt message
                if self.__aes_cipher is not None:
                    data = self.__aes_cipher.add_padding(data) # -> It cant broke the model because it fills all the block space
                    data = self.__aes_cipher.encrypt(data)
                # Send message
                self.__socket_lock.acquire()
                self.__socket.sendall(data)
                self.__socket_lock.release()
                return True
            
            except Exception as e:
                # If some error happened, the connection is closed
                self.__terminate_flag.set() 
                return False
        else:
            return False

    ###########################
    #    Command Callbacks    #
    ###########################

    def notify_heartbeat(self,node):
        """
        Notify that a heartbeat was received.
        """
        self.notify(Events.BEAT_RECEIVED_EVENT, node)

    def notify_conn_to(self, h, p):
        """
        Notify to the parent node that `CONN_TO` has been received.
        """
        self.notify(Events.CONN_TO_EVENT, (h,p))

    def notify_start_learning(self, r, e):
        """
        Notify to the parent node that `START_LEARNING_EVENT` has been received.
        """
        self.notify(Events.START_LEARNING_EVENT, (r,e))

    def notify_stop_learning(self,cmd):
        """
        Notify to the parent node that `START_LEARNING_EVENT` has been received.
        """
        self.notify(Events.STOP_LEARNING_EVENT, None)

    def notify_params(self,params):
        """
        Notify to the parent node that `PARAMS` has been received.
        """
        self.notify(Events.PARAMS_RECEIVED_EVENT, (params))

    def notify_metrics(self,node,round,loss,metric):
        """
        Notify to the parent node that `METRICS` has been received.
        """
        self.notify(Events.METRICS_RECEIVED_EVENT, (node, round, loss, metric))

    def notify_train_set_votes(self,node,votes):
        """
        Set the last ready round of the other node.

        Args:
            round: The last ready round of the other node.
        """
        self.notify(Events.TRAIN_SET_VOTE_RECEIVED_EVENT, (node,votes))